searchData={"items":[{"type":"module","title":"Hardhat","doc":"An opinionated, production-ready HTTP client for Elixir services. 👷🌐","ref":"Hardhat.html"},{"type":"module","title":"What's included - Hardhat","doc":"- Connection pooling per-client module\n- Integration with `telemetry` and `opentelemetry` instrumentation\n- Circuit breaking for repeatedly failed requests\n- Automatic retries for failed requests\n- Timeout and `deadline` support","ref":"Hardhat.html#module-what-s-included"},{"type":"module","title":"Why Hardhat? - Hardhat","doc":"In 2021, my employer was in the process of [refactoring its monolithic Phoenix application into a small number of decoupled services](https://www.youtube.com/watch?v=Py8WK4rBNqQ), so we needed better reliability and observability at the boundaries of our services. We had experienced multiple production incidents related to exhaustion of a single, shared connection pool for outgoing HTTP requests. Additionally, we had built a number of custom clients for external SaaS APIs but had no consistency between them.\n\nI set out to address these problems by creating a standard HTTP client library, upon which individual teams could build clients for internal and external APIs and get reliability and observability, relatively for-free. `Hardhat` was born (its name comes from \"hardened HTTP client\", and that you should wear a hardhat to protect your head in dangerous construction areas).\n\n`Hardhat` attempts to walk the line of baking-in sensible defaults so the upfront effort is minimal, but also allowing you to customize and extend almost every part of the built-in functionality. It is not a low-level HTTP client, but adds functionality on top of `Tesla` and `Finch`, and draws upon well-crafted libraries like `:opentelemetry`, `:telemetry`, `:fuse`, `Regulator`, and `Deadline`.\n\nRegrettably, my employer did not see fit to release `Hardhat` as open-source software, so this library recreates it from scratch, built only from my own recollections and the help of the community.","ref":"Hardhat.html#module-why-hardhat"},{"type":"module","title":"Installation - Hardhat","doc":"Add `hardhat` to the dependencies in your `mix.exs`:\n\n```elixir\n  def deps do\n    [\n      {:hardhat, \"~> 1.0.0\"}\n    ]\n  end\n```","ref":"Hardhat.html#module-installation"},{"type":"module","title":"Getting started - Hardhat","doc":"`Hardhat` is designed to be easy for creating quick wrappers around HTTP APIs,\nbut includes many options for customization. To define a simple client, do something like the following:\n\n```elixir\n# Define a client module:\ndefmodule SocialMediaAPI do\n  use Hardhat\nend\n\n# Add it to your supervisor (required):\ndefmodule MyApp.Sup do\n  use Supervisor\n\n  def start_link(init_arg) do\n    Supervisor.start_link(__MODULE__, init_arg, name: __MODULE__)\n  end\n\n  @impl true\n  def init(_init_arg) do\n    children = [\n      SocialMediaAPI\n    ]\n\n    Supervisor.init(children, strategy: :one_for_one)\n  end\nend\n\n# Use your client to make requests:\nSocialMediaAPI.get(\"http://media-api.social/posts\")\n```\n\nAs mentioned in the example above, it is **imperative** for you to supervise the client module that includes the `use Hardhat` macro. Without starting the client under supervision, you will not be able to make requests. See [Connection pools](#module-connection-pools) below for more information.","ref":"Hardhat.html#module-getting-started"},{"type":"module","title":"General behavior - Hardhat","doc":"`Hardhat` is built on top of `Tesla`, and uses `Finch` as the adapter. Because\n`Tesla` is the foundation, you are welcome to use publicly available\n`Tesla.Middleware` modules in your `Hardhat`-based client (with the exception\nthat [we recommend](#module-timeouts-and-deadlines) you use\n`Hardhat.Middleware.Timeout` instead of `Tesla.Middleware.Timeout`).\n\n```elixir\ndefmodule SomeJSONAPI do\n  use Hardhat\n\n  plug Tesla.Middleware.BaseUrl, \"https://my-json.api/\"\n  plug Tesla.Middleware.JSON\nend\n```\n\nIn addition to the adapter selection and default `Tesla` behavior,\n`use Hardhat` will inject the common functionality [listed above](#module-what-s-included) *after* any middleware that you supply via [`plug`](`Tesla.Builder.plug/2`). The current list is as follows:\n\n* `Hardhat.Middleware.DeadlinePropagation` ([see below](#module-deadline-propagation))\n* `Tesla.Middleware.Retry` ([see below](#module-retries))\n* Either `Tesla.Middleware.Fuse` or `Hardhat.Middleware.Regulator` ([see below](#module-failure-detection))\n* `Tesla.Middleware.Telemetry` ([see below](#module-telemetry-and-tracing))\n* `Tesla.Middleware.OpenTelemetry` ([see below](#module-telemetry-and-tracing))\n* `Hardhat.Middleware.PathParams`\n\nEach of the included middlewares that have configuration have defaults defined by functions in `Hardhat.Defaults` and can be customized by defining a function of the same name in your client module. Inside those functions you can set your own static defaults or get runtime configuration using `Application.get_env/3`. The [options](`t:Keyword.t/0`) you return will be merged with the defaults when the middleware is invoked. Examples of this pattern are in each of the sections below.","ref":"Hardhat.html#module-general-behavior"},{"type":"module","title":"Connection pools - Hardhat","doc":"As mentioned above, `Hardhat` uses `Finch` as the adapter. By [default](`Hardhat.Defaults.pool_configuration/1`), `Hardhat` specifies a connection pool of size `10` but sets no [other options](`Finch.start_link/1`) on the adapter. The name of the `Finch` process is proscribed by the `use Hardhat` macro, but you can set any other options for the pool that you like, including creating more than one pool or setting the HTTP protocol or TLS options by overriding the `pool_configuration/1` function.\n\n```elixir\ndefmodule H2Client do\n  use Hardhat\n\n  # This function overrides the configuration coming from `Hardhat.Defaults`.\n  # The `overrides` will be passed from your process supervision initial\n  # arguments.\n  def pool_configuration(_overrides \\\\ %{}) do\n    %{\n      # By default we'll use HTTP/2, with 3 pools of one connection each\n      :default => [\n        protocols: [:http2],\n        count: 3\n      ],\n      # For this host only, we're using HTTP/1.1 and a single pool of 20\n      # connections\n      \"https://some-http1-only-host.com/\" => [\n        size: 20\n      ]\n    }\n  end\nend\n```","ref":"Hardhat.html#module-connection-pools"},{"type":"module","title":"Telemetry and tracing - Hardhat","doc":"`Hardhat` includes the stock `Tesla.Middleware.Telemetry` for injecting your own metrics and monitoring systems into its operation. The events emitted by this middleware are:\n\n* `[:tesla, :request, :start]` - at the beginning of the request\n* `[:tesla, :request, :stop]` - at the completion of the request\n* `[:tesla, :request, :exception]` - when a non-HTTP-status exception occurs\n\n```elixir\ndefmodule TelemetryClient do\n  use Hardhat\nend\n\n:telemetry.attach(\n  \"my handler\",\n  [:tesla, :request, :stop],\n  fn _event, measures, _metadata, _config ->\n     # Don't do this, attach to your metrics system instead\n     IO.puts(\"Made a request in #{measures.duration}\")\n  end,\n  nil\n)\n```\n\n`Hardhat` wraps each request in an [OpenTelemetry](`Tesla.Middleware.OpenTelemetry`) span and propagates the trace context to the destination host. It will observe any [path parameters](`Hardhat.Middleware.PathParams`) you interpolate into the URL so that similar spans can be easily aggregated. To disable propagation or change other behavior of the tracing, override the `opentelemetry_opts/0` function:\n\n```elixir\ndefmodule NoPropagationClient do\n  use Hardhat\n\n  def opentelemety_opts do\n    [\n      # Disable trace propagation for talking to a third-party API\n      propagator: :none\n    ]\n  end\nend\n```\n\nFor more information about the options available to tracing, see `Hardhat.Defaults.opentelemetry_opts/0` or `Tesla.Middleware.OpenTelemetry`.","ref":"Hardhat.html#module-telemetry-and-tracing"},{"type":"module","title":"Failure detection - Hardhat","doc":"`Hardhat` provides two different failure detection and backoff strategies:\n\n* Static circuit breaking with `:fuse` (`Tesla.Middleware.Fuse`)\n* Dynamic request rate regulation ([AIMD](`Regulator.Limit.AIMD`)) with `Regulator` (via `Hardhat.Middleware.Regulator`)\n\nThese strategies cannot be used together safely, so you must choose one when defining your client. If your needs are simple and hard failures are relatively rare, `:fuse` is an easier strategy to comprehend and implement because it uses a straightforward failure-counting algorithm, and completely turns off requests when the configured threshold is reached. If you have a more complicated setup or high traffic, and do not want to spend as much time tuning your failure behavior, the `:regulator` strategy might be for you. `Regulator` allows your client to adapt to rapidly changing conditions by reducing the amount of concurrent work in the presence of failure, without causing a hard stop to activity. On the other hand, if your concurrent utilization is low, it might also bound your maximum concurrency even when requests are not failing.\n\n#","ref":"Hardhat.html#module-failure-detection"},{"type":"module","title":"The Fuse strategy - Hardhat","doc":"The `:fuse` failure detection strategy is configured with two functions in your client which have default implementations that are injected at compile-time:\n\n* [`fuse_opts()`](`Hardhat.Defaults.fuse_opts/1`) - configuration for the middleware\n* [`should_melt(result)`](`Hardhat.Defaults.should_melt/1`) - whether the result of the request is considered a failure\n\nYou can override their default behavior by redefining the functions:\n\n```elixir\n# This module uses `:fuse` for failure detection and backoff.\ndefmodule HardCutoffClient do\n  use Hardhat # defaults to `:fuse`\n\n  # This is also valid:\n  # use Hardhat, strategy: :fuse\n\n  # Customize fuse's circuit-breaking behavior\n  def fuse_opts do\n    [\n      # 10 failed requests in 0.5sec flips the breaker, which resets after 1sec\n      opts: {{:standard, 10, 500}, {:reset, 1_000}}\n    ]\n  end\n\n  # Customize how responses are determined to be failures,\n  # in this case only TCP/adapter-type errors are considered\n  # failures, any valid response is fine.\n  def should_melt(result) do\n    case result do\n      {:error, _} -> true\n      {:ok, %Tesla.Env{}} -> false\n    end\n  end\nend\n```\n\n#","ref":"Hardhat.html#module-the-fuse-strategy"},{"type":"module","title":"The Regulator strategy - Hardhat","doc":"The `:regulator` failure detection strategy is configured with two functions in your client which have default implementations that are injected at compile-time:\n\n* [`regulator_opts()`](`Hardhat.Defaults.regulator_opts/1`) - configuration for the middleware\n* [`should_regulate(result)`](`Hardhat.Defaults.should_regulate/1`) - whether the result of the request is considered a failure\n\nYou can override their default behavior by redefining the functions:\n\n```elixir\n# This module uses `Regulator` for failure detection and backoff\ndefmodule DynamicRegulationClient do\n  use Hardhat, strategy: :regulator # overrides default of `:fuse`\n\n  # Customize Regulator's ramp-up and backoff strategy\n  def regulator_opts do\n    [\n      # Backoff on failure by half instead of 90%\n      backoff_ratio: 0.5\n    ]\n  end\n\n  # Customize how responses are determined to be failures,\n  # in this case TCP/adapter-level errors are considered failures,\n  # as well as HTTP `429 Too Many Requests` responses.\n  def should_regulate(result) do\n    case result do\n      {:error, _} -> true\n      {:ok, %Tesla.Env{status: 429}} -> true\n      {:ok, %Tesla.Env{}} -> false\n    end\n  end\nend\n```\n\n#","ref":"Hardhat.html#module-the-regulator-strategy"},{"type":"module","title":"Disabling failure detection - Hardhat","doc":"> ##","ref":"Hardhat.html#module-disabling-failure-detection"},{"type":"module","title":"Warning {: .warning} - Hardhat","doc":"> We do not recommend disabling failure detection and backoff strategies because they expose you to encountering cascading failure and slowdown when the target service or network is encountering issues.\n\nIf you want to disable failure detection, set the strategy to `:none`:\n\n```elixir\ndefmodule WildAndFree do\n  use Hardhat, strategy: :none\nend\n```","ref":"Hardhat.html#module-warning-warning"},{"type":"module","title":"Retries - Hardhat","doc":"`Hardhat` injects automatic retries on your requests using `Tesla.Middleware.Retry`. Retries are configured with two functions in your client which have default implementations that are injected at compile-time:\n\n* [`retry_opts()`](`Hardhat.Defaults.retry_opts/1`) - configuration for the middleware\n* [`should_retry(result)`](`Hardhat.Defaults.should_retry/1`) - whether the result of the request can be retried\n\nYou can override their default behavior by redefining the functions:\n\n```elixir\n# This client retries requests\ndefmodule SomeRetriesClient do\n  use Hardhat\n\n  def retry_opts do\n    [\n      # Retry up to 5 times\n      max_retries: 5,\n      # Delay at least 75ms between attempts\n      delay: 75,\n      # Delay at most 500ms between any attempts\n      max_delay: 500\n    ]\n  end\nend\n\n# This client disables retries entirely!\ndefmodule NoRetriesClient do\n  use Hardhat\n\n  # Override should_retry to disable retries\n  def should_retry(_), do: false\nend\n```\n\n> #","ref":"Hardhat.html#module-retries"},{"type":"module","title":"Interaction with failure detection {: .warning} - Hardhat","doc":"> Retries can interact very negatively with [failure detection](#module-failure-detection), potentially triggering backoff behavior too quickly. Be sure to avoid retrying when the failure detector returns `{:error, :unavailable}`, which indicates that the circuit breaker has blown in the `:fuse` strategy, or the limiter is out of capacity in the `:regulator` strategy.\n>\n> The default implementation of `should_retry/1` implements this behavior.","ref":"Hardhat.html#module-interaction-with-failure-detection-warning"},{"type":"module","title":"Timeouts and deadlines - Hardhat","doc":"Timeouts are an essential liveness technique that help prevent your application from waiting a long time for a slow response to come back from a request. Deadlines extend the timeout pattern to cover entire workflows spanning multiple services, and ensure responsiveness for requests at the edge of your infrastructure while avoiding doing work that will not complete in a timely fashion.\n\nHardhat supports both individual timeouts and global deadlines via its custom `Hardhat.Middleware.Timeout` middleware, but it is **not** included in the default middleware stack.\n\nWhen a timeout value is specified in the middleware options, the lesser of that value and any active deadline will be used for the timeout duration. If the configured timeout is shorter than the active deadline, the deadline propagated downstream will be set to the shorter timeout value.\n\n```elixir\ndefmodule TimeoutClient do\n  use Hardhat\n\n  # Requests will time out after 100ms\n  plug Hardhat.Middleware.Timeout, timeout: 100\nend\n\n# Set a deadline of 50ms\nDeadline.set(50)\n\n# This will timeout after 50ms instead of the configured 100ms\nTimeoutClient.get(\"http://google.com\")\n\n# Set a deadline of 500ms\nDeadline.set(500)\n\n# This will timeout after 100ms (not 500ms), propagating a deadline of 100ms\nTimeoutClient.get(\"http://elixir-lang.org\")\n```\n\nWhen a timeout occurs, a `:timeout_exceeded` event will be added to the current `OpenTelemetry` span.\n\n> #","ref":"Hardhat.html#module-timeouts-and-deadlines"},{"type":"module","title":"Using `Tesla.Middleware.Timeout` instead {: .warning} - Hardhat","doc":"> Because implementing timeouts requires spawning a process that carries out the rest of the request, we recommend using `Hardhat`'s bundled timeout middleware. If you use the standard middleware bundled with `Tesla`, you must propagate `OpenTelemetry` context and `Deadline` information yourself.\n\n#","ref":"Hardhat.html#module-using-tesla-middleware-timeout-instead-warning"},{"type":"module","title":"Deadline propagation - Hardhat","doc":"The default middleware stack will propagate any `Deadline` you have set for the current process, regardless of whether you are using the `Hardhat.Middleware.Timeout` middleware in your client. The propagation consists of a request header (default `\"deadline\"`) whose value is the current deadline as an integer in milliseconds. To change the header name, override the [`deadline_propagation_opts`](`Hardhat.Defaults.deadline_propagation_opts/0`) callback:\n\n```elixir\ndefmodule CustomPropagationClient do\n  use Hardhat\n\n  def deadline_propagation_opts do\n    [\n      header: \"countdown-expires-in\"\n    ]\n  end\nend\n```","ref":"Hardhat.html#module-deadline-propagation"},{"type":"module","title":"Testing - Hardhat","doc":"Testing HTTP clients can be tricky, partly because they are software designed to interact with the outside world. Here are the primary strategies that one can take when testing `Hardhat` clients:\n\n* [`Mox`](https://hexdocs.pm/mox/Mox.html), which can generate a mock `Tesla.Adapter`.\n* [`Bypass`](https://hexdocs.pm/bypass/Bypass.html), which runs a `Plug` web server to handle requests from your client.\n* [`ExVCR`](https://hexdocs.pm/exvcr/readme.html), which replaces the adapter-level library with a double that records and replays responses.\n\nWe do not recommend using `Tesla.Mock` for testing. Any of the three options above have superior behavior under complicated testing conditions, including spawning child processes via timeouts.\n\n### `Mox`\n\n`Mox` allows us to define a custom `Tesla.Adapter` for use only in tests. First, we need to generate the mock adapter (put this in `test_helper.exs`):\n\n```elixir\nMox.defmock(MockAdapter, for: Tesla.Adapter)\n```\n\nThen configure your `Hardhat`-based client to use this adapter in `config/test.exs`:\n\n```elixir\nimport Config\n\nconfig :tesla, MyHardhatClient, adapter: MockAdapter\n```\n\nThen in your tests, set expectations on the adapter:\n\n```elixir\ndefmodule MyHardhatClient.Test do\n  use ExUnit.Case, async: true\n  import Mox\n\n  # Checks your mock expectations on each test\n  setup :verify_on_exit!\n\n  test \"it works\" do\n    expect(MockAdapter, :call, fn env, opts -> {:ok, %{env | status: 204}} end)\n\n    assert {:ok, %Tesla.Env{status: 204}} = MyHardhatClient.get(\"https://foo.bar/\")\n  end\nend\n```\n\nThis setup will work even if you are using `Hardhat.Middleware.Timeout` in your middleware, as child processes automatically inherit expectations and stubs defined by `Mox`.\n\n### `Bypass`\n\n`Bypass` starts a web server in a new process that handles requests from your client. In order to use it effectively in tests, you will need to be able to set the hostname for each request (or for the current process), which might be challenging if you are already using `Tesla.Middleware.BaseUrl` in your client. One strategy using the process dictionary is shown below:\n\n```elixir\ndefmodule ClientWithBypassUrl do\n  use Hardhat\n\n  plug Tesla.Middleware.BaseUrl, base_url()\n\n  def base_url do\n    Process.get(:bypass_url) ||\n      Application.get_env(:my_app, __MODULE__, [])[:base_url]\n  end\nend\n```\n\nAnd then in the test:\n\n```elixir\ndefmodule ClientWithBypassUrl.Test do\n  use ExUnit.Case, async: true\n\n  setup do\n    bypass = Bypass.open()\n    Process.put(:bypass_url, \"http://localhost:#{bypass.port}\")\n    {:ok, %{bypass: bypass}}\n  end\n\n  test \"works with bypass\", %{bypass: bypass} do\n    Bypass.expect(bypass, fn conn -> Plug.Conn.resp(conn, 204, \"\") end)\n\n    assert {:ok, %Tesla.Env{status: 204}} = ClientWithBypassUrl.get(\"/test\")\n  end\nend\n```\n\n### `ExVCR`\n\n[`exvcr`](https://hexdocs.pm/exvcr) intercepts calls into specific HTTP client libraries (like `Finch`) and returns pre-determined responses. Developers can execute tests in a recording mode, which will initialize the \"cassettes\" by executing real requests and recording the real responses into JSON files on disk. Once recorded, a call to `use_cassette` inside the test selects a particular session for replay.\n\n```elixir\ndefmodule PreRecordedClient.Test do\n  use ExUnit.Case, async: true\n  # Be sure to set Finch as the adapter in this call, or whatever you configured\n  # your Hardhat client to use\n  use ExVCR.Mock, adapter: ExVCR.Adapter.Finch\n\n  test \"returns a recorded response\" do\n    use_cassette \"example\" do\n      assert {:ok, %Tesla.Env{status: 204}} = PreRecordedClient.get(\"/\")\n    end\n  end\nend\n```","ref":"Hardhat.html#module-testing"},{"type":"module","title":"Hardhat.Defaults","doc":"Contains default implementations of functions that can be overridden in\nclients that `use Hardhat`.","ref":"Hardhat.Defaults.html"},{"type":"function","title":"Hardhat.Defaults.deadline_propagation_opts/0","doc":"Default implementation of the options to `Hardhat.Middleware.DeadlinePropagation` that\nis included in the default middleware stack.\n\nThese defaults include:\n- `:header` - `\"deadline\"` is the name of the header added to the request","ref":"Hardhat.Defaults.html#deadline_propagation_opts/0"},{"type":"function","title":"Hardhat.Defaults.fuse_opts/1","doc":"Default implementation of the options to the `Tesla.Middleware.Fuse` that is\nincluded in the default middleware stack.\n\nTakes the client module name as an argument.\n\nThese defaults include:\n- `:opts` - The fuse will blow after 50 errors with 1 second, and reset after 2 seconds\n- `:keep_original_error` - The original error will be returned when a fuse first blows\n- `:should_melt` - The `should_melt/1` function defined in the client module is used\n   (by default this is `Hardhat.Defaults.should_melt/1`)\n- `:mode` - The fuse uses `:async_dirty` mode to check the failure rate, which may result\n  in some delay in blowing the fuse under high concurrency, but it will not serialize\n  checks to the fuse state through a single process\n\nSee `Tesla.Middleware.Fuse` for more information on how to configure the middleware.","ref":"Hardhat.Defaults.html#fuse_opts/1"},{"type":"function","title":"Hardhat.Defaults.opentelemetry_opts/0","doc":"Default options for the `Tesla.Middleware.OpenTelemetry` middleware.\n\nThe options include:\n* `:span_name` - override span name. Can be a `String` for a static span name,\n   or a function that takes the `Tesla.Env` and returns a `String`. The\n   default span name is chosen by the middleware.\n* `:propagator` - configures trace headers propagators. Setting it to `:none`\n   disables propagation. Any module that implements `:otel_propagator_text_map`\n   can be used. Defaults to calling `:opentelemetry.get_text_map_injector/0`\n* `:mark_status_ok` - configures spans with a list of expected HTTP error codes to be\n   marked as ok, not as an error-containing spans. The default is empty.","ref":"Hardhat.Defaults.html#opentelemetry_opts/0"},{"type":"function","title":"Hardhat.Defaults.pool_configuration/1","doc":"The default configuration for the connection pool(s) that will be created when your\nclient is added to the supervision tree. Overrides to the pool can be passed\nat startup time as an argument in the supervision tree (see `Supervisor.child_spec/2`).\n\nThis creates a connection pool of size `10`. See `Finch.start_link/1` for more details.","ref":"Hardhat.Defaults.html#pool_configuration/1"},{"type":"function","title":"Hardhat.Defaults.regulator_opts/1","doc":"Default options for the `Regulator` middleware, which can be used as an\nalternative circuit-breaking strategy to `:fuse`.\n\nThe defaults include:\n* `:min_limit` - The minimum concurrency limit (defaults to 5)\n* `:initial_limit` - The initial concurrency limit when the regulator is installed (deafults to 20)\n* `:max_limit` - The maximum concurrency limit (defaults to 200)\n* `:step_increase` - The number of tokens to add when regulator is increasing the concurrency limit (defaults to 10).\n* `:backoff_ratio` - Floating point value for how quickly to reduce the concurrency limit (defaults to 0.9)\n* `:target_avg_latency` - This is the average latency in milliseconds for the system regulator is protecting. If the average latency drifts above this value Regulator considers it an error and backs off. Defaults to 5.\n* `:should_regulate` - Whether to consider the result of the request as failed, defaults to `should_regulate/1`.","ref":"Hardhat.Defaults.html#regulator_opts/1"},{"type":"function","title":"Hardhat.Defaults.retry_opts/1","doc":"Default implementation of the options to `Tesla.Middleware.Retry` that is\nincluded in the default middleware stack.\n\nThese defaults include:\n- `:delay` - The base delay in milliseconds (positive integer, defaults to 50)\n- `:max_retries` - maximum number of retries (non-negative integer, defaults to 3)\n- `:max_delay` - maximum delay in milliseconds (positive integer, defaults to 300)\n- `:should_retry` - function to determine if request should be retried, defaults to `should_retry/1`\n- `:jitter_factor` - additive noise proportionality constant (float between 0 and 1, defaults to 0.2)","ref":"Hardhat.Defaults.html#retry_opts/1"},{"type":"function","title":"Hardhat.Defaults.should_melt/1","doc":"Default implementation of the \"melt test\" for circuit breaking. This\nfunction will cause the circuit breaker to record an error when the\nresult of a request is:\n\n* A TCP-level error, e.g. `{:error, :econnrefused}`\n* An HTTP status that indicates a server error or proxy-level error (>= 500)\n* An `429 Too Many Requests` HTTP status","ref":"Hardhat.Defaults.html#should_melt/1"},{"type":"function","title":"Hardhat.Defaults.should_regulate/1","doc":"Default implementation of the \"failure test\" for dynamic regulation. This\nfunction will cause the `Regulator` to record an error when the\nresult of a request is:\n\n* A TCP-level error, e.g. `{:error, :econnrefused}`\n* An HTTP status that indicates a server error or proxy-level error (>= 500)\n* An `429 Too Many Requests` HTTP status","ref":"Hardhat.Defaults.html#should_regulate/1"},{"type":"function","title":"Hardhat.Defaults.should_retry/1","doc":"Default implementation of the \"retry test\" for retries. This\nfunction will cause requests to be retried when the result of the\nrequest is:\n\n* A TCP-level error, e.g. `{:error, :econnrefused}`\n* An HTTP status that indicates a server error or proxy-level error (>= 500)\n* An `429 Too Many Requests` HTTP status\n\nIn the case where the circuit breaker has been triggered, or the request method\nwas `POST`, requests will not be retried.","ref":"Hardhat.Defaults.html#should_retry/1"},{"type":"module","title":"Hardhat.Middleware.DeadlinePropagation","doc":"Propagates `Deadline` information to the server being called via\nan HTTP header. _This middleware is part of the default stack_.\n\nExpects a `:header` option, which defaults to `\"deadline\"`. See `Hardhat.Defaults.deadline_propagation_opts/0`.","ref":"Hardhat.Middleware.DeadlinePropagation.html"},{"type":"module","title":"Hardhat.Middleware.PathParams","doc":"Use templated URLs with separate params. Unlike `Tesla.Middleware.PathParams`,\nwe ensure that all parameters are URL-safe. _This middleware is part of the\ndefault stack_.","ref":"Hardhat.Middleware.PathParams.html"},{"type":"module","title":"Hardhat.Middleware.Regulator","doc":"Limits concurrent requests to the server based on a [loss based dynamic limit algorithm](`Regulator.Limit.AIMD`).\nAdditively increases the concurrency limit when there are no errors and multiplicatively\ndecrements the limit when there are errors.\n\n_This middleware is part of the default stack when giving the `strategy: :regulator` option to `use Hardhat`._\n\nSee also:\n- Configuration options: `Hardhat.Defaults.regulator_opts/1`\n- Determining what amounts to an error: `Hardhat.Defaults.should_regulate/1`","ref":"Hardhat.Middleware.Regulator.html"},{"type":"module","title":"Hardhat.Middleware.Timeout","doc":"Abort HTTP requests after the given timeout or the current `Deadline`,\nand emit `OpenTelemetry` trace events on timeouts.\n\nOptions:\n- `:timeout` - (required) timeout in milliseconds","ref":"Hardhat.Middleware.Timeout.html"},{"type":"module","title":"OpenTelemetry - Hardhat.Middleware.Timeout","doc":"Implementing a timeout necessitates moving the request into a new process,\nand then waiting on that new process's completion (or aborting after the timeout).\nSince `OpenTelemetry` tracing context is stored in the process dictionary, that\ncontext must be explicitly propagated to the new process. This middleware uses\n`OpentelemetryProcessPropagator` for this purpose.\n\nIn the event of a timeout result in this middleware, a new `timeout_exceeded` event\nwill be added to the trace. The event will include these attributes:\n\n- `module` - the client module that includes this middleware\n- `timeout` - the duration that was exceeded","ref":"Hardhat.Middleware.Timeout.html#module-opentelemetry"},{"type":"module","title":"Deadline - Hardhat.Middleware.Timeout","doc":"When the caller has set a `Deadline` for the current process, that limit will be\nrespected by this middleware. The effective timeout chosen will be the **lesser** of the time\nremaining on the current deadline and the duration given in the `:timeout` option.","ref":"Hardhat.Middleware.Timeout.html#module-deadline"},{"type":"extras","title":"Changelog","doc":"# Changelog","ref":"changelog.html"},{"type":"extras","title":"v1.0.2 - Changelog","doc":"* Bump ex_doc from 0.30.9 to 0.31.0 by [@dependabot](https://github.com/dependabot) ([#14](https://github.com/seancribbs/hardhat/pull/14))\n* Bump dialyxir from 1.4.2 to 1.4.3 by [@dependabot](https://github.com/dependabot) ([#15](https://github.com/seancribbs/hardhat/pull/15))\n* Bump ex_doc from 0.31.0 to 0.31.1 by [@dependabot](https://github.com/dependabot) ([#17](https://github.com/seancribbs/hardhat/pull/17))\n* Bump finch from 0.16.0 to 0.17.0 [@dependabot](https://github.com/dependabot) ([#16](https://github.com/seancribbs/hardhat/pull/16))\n* Update documentation for changed Finch pool options by [@seancribbs](https://github.com/seancribbs)","ref":"changelog.html#v1-0-2"},{"type":"extras","title":"v1.0.1 - Changelog","doc":"* Small stylistic improvements by [@hissssst](https://github.com/hissssst) ([#12](https://github.com/seancribbs/hardhat/pull/12))\n* Bump `regulator` from 0.5.0 to 0.6.0 by [@dependabot](https://github.com/dependabot) ([#13](https://github.com/seancribbs/hardhat/pull/13))","ref":"changelog.html#v1-0-1"},{"type":"extras","title":"v1.0.0 - Changelog","doc":"Initial release!","ref":"changelog.html#v1-0-0"}],"content_type":"text/markdown"}